/*----------------------------------------------------------------------
* File    : xminer.cpp
* Contents: XMiner algorithm for finding frequent representative
*           itemsets in incomplete datasets
* Author  : Michael Mampaey
* Date    : 18/05/2006
* --------------------------------------------------------------------*/

#include "data.h"
#include "item.h"
#include "xminer.h"
#include <iostream>
#include <map>
#include <time.h>
using namespace std;


XMiner::XMiner() : data(0), out(0), verbose(false)
{
    
}


XMiner::~XMiner() 
{
    if (data) delete data;
    if (out) fclose(out);
}


double XMiner::mine()
{
	clock_t start = clock();
    
    if (verbose) {
        if (minrep<0||minrep>1)
            cout << "WARNING: minrep is not in [0.0;1.0]" << endl;
        if (minsup<0||minsup>1)
            cout << "WARNING: minsup is not in [0.0;1.0]" << endl;
    }
    
    int tnr = 0;
    int attrcount = 0;
    set<Attribute> root;
    multiset<Attribute> *allitems = new multiset<Attribute>;
    set<Attribute>::iterator ita;
    
    // Read data
    if (verbose) cout << "Reading data... " << flush;
    
    while(Transaction *t = data->getNext()) {
        // commented lines inhere are for general itemsets
        for(int i=0; i<t->length; ++i) {
            Value item(0,0);
            //Value starToken(Value::starToken,0);
            //Value questToken(Value::questToken,0);
            Attribute a(i);
            a.attrItem->support = i;
            ita = root.insert(a).first;
            
            if (!t->t[i]->empty() && (*(t->t[i]))[0]!='?') {
                if (item.setVal(t->t[i]))
                    t->t[i] = 0; // pointer stealing -- no double delete
                ita->addValue(item)->transactions->push_back(tnr);
                ita->attrItem->transactions->push_back(tnr);
                //ita->addValue(starToken)->transactions->push_back(tnr);
            }
            else {
                //ita->addValue(questToken)->transactions->push_back(tnr);
            }
        }
        ++tnr;
        delete t;
    }
    
    if (verbose) cout << "Done." << endl;
    
    tcount = tnr;
    attrcount = root.size();
    double minsuprepcount = minsup*minrep*tcount;
    minrepcount = (int)(minrep*tcount);
    candidates = 0;
    extensibleCandidates = 0;
    
    if (minrepcount<1 || minsup*tcount<1) {
        cerr<<"ERROR: minsup and/or minrep too low for database."<<endl;
        return -1;
    }
    
    // Attribute items (Representativity)
    if (verbose) cout << "Pruning infrequent items... " << flush;
    
    while((ita=root.begin()) != root.end()) {
        
        ita->attrItem->support = ita->attrItem->transactions->size();
        if ((double)(ita->attrItem->support)>=minrepcount) {
            Attribute a(ita->attr);
            a.stealAttrItem(*ita);
            ++candidates;
            
            set<Value>::iterator itv;
            while((itv=ita->values->begin()) != ita->values->end()) {
                itv->support = itv->transactions->size();
                if (itv->val==Value::starToken || itv->val==Value::questToken) {
                    if (itv->support>=minrepcount) {
                        a.addValue(*itv);
                        itv->transactions = 0;
                    }
                }
                else {
                    if ((double)(itv->support)>=minsuprepcount) {
                        a.addValue(*itv);
                        itv->transactions = 0;
                    }
                }
                ita->values->erase(itv);
            }
            if (!(a.values->empty())) {
                a.reset();
                allitems->insert(a);
                candidates += a.values->size();
            }
        }
        root.erase(ita);
    }
    
    if (verbose) cout << "Done." << endl;
    
    if (out) fprintf(out,"(%d/%d)\n",tnr,tnr);
    
    pair<int,string*> *itemset = new pair<int,string*>[allitems->size()];
    pair<int,string*> *comb = new pair<int,string*>[allitems->size()];
    
    if (verbose) cout << "Starting the mining... " << flush;
    double added = grow(allitems, itemset, 1, comb, 0);
    
    if (verbose) {
    	cout << "Done." << endl;
    	cout << "Summary:" << endl;
        cout << "Number of transactions:\t\t\t" << tnr << endl;
        cout << "Number of attributes:\t\t\t" << attrcount << endl;
        cout << "Candidate itemsets:\t\t\t" << candidates+1 << endl;
        cout << "Extensible candidate itemsets:\t\t" << extensibleCandidates+1 << endl;
        cout << "Frequent representative itemsets:\t" << added+1 << endl;
        cout << "[" << (clock()-start)/double(CLOCKS_PER_SEC) << "sec]" << endl;
    }
    
    delete [] comb;
    delete [] itemset;
    delete allitems;
    
    return added+1;
}



double XMiner::grow(multiset<Attribute> *items, pair<int,string*> *itemset, int depth, pair<int,string*> *comb, int cl)
{
    double added = 0;
    multiset<Attribute>::const_iterator it1, it2, begin, end;
    
    it1 = items->end();
    
    while(it1!=items->begin()) { // rightmost DFS
        it1--;
        Value& v1s = *(it1->attrItem);
        
        map<int,bool> fr;
        map<int,Attribute*> fa;
        
        /* *** Attribute Items (Representativity) -- Set theory! *** */
        Value tail(Value::star);
        tail.support = v1s.support;
        if (depth==1) {
            *(tail.transactions) = *(v1s.transactions);
        }
        
        map<int, double> maxtails;
    
        double tmptail = tail.support;
      
        for(it2=items->end(), it2--; it2!=it1; it2--) { //reverse as values
            
            Value& v2s = *(it2->attrItem);
            
            fa.insert(pair<int, Attribute*>(it2->attr, new Attribute(it2->attr)));
            Attribute& a = *(fa[it2->attr]);
            
            insert_iterator< vector<int> > res_ins(*(a.attrItem->transactions), a.attrItem->transactions->begin());
            if (depth==1) 
                set_difference(v1s.transactions->begin(), v1s.transactions->end(), v2s.transactions->begin(), v2s.transactions->end(), res_ins);
            else
                set_difference(v2s.transactions->begin(), v2s.transactions->end(), v1s.transactions->begin(), v1s.transactions->end(), res_ins);
            a.attrItem->support = v1s.support - a.attrItem->transactions->size();
            
        if (fr[it2->attr]=((double)(a.attrItem->support)>=minrepcount)) {
                
        if ((double)(tail.support)<minrepcount) {
                    maxtails.insert(pair<int,double>(it2->attr, tmptail));
        }
        else {
                    Value tmp = tail;
                    tail.transactions->clear();
                    insert_iterator< vector<int> > res_ins(*(tail.transactions), tail.transactions->begin());
                    
                    if (depth==1) {
                        set_intersection(tmp.transactions->begin(), tmp.transactions->end(), v2s.transactions->begin(), v2s.transactions->end(), res_ins);
                        tail.support = tail.transactions->size();
                    }
                    else {
                        set_union(a.attrItem->transactions->begin(), a.attrItem->transactions->end(), tmp.transactions->begin(), tmp.transactions->end(), res_ins);
                        tail.support = v1s.support - tail.transactions->size();
                    }
                    maxtails.insert(pair<int,double>(it2->attr, tail.support));
                    tmptail = tail.support;
               }
            }
        
        }
        
        /* *** Attribute values (support and extensibility) *** */
        it1->reset();
        while(it1->currentIsValid()) {
            const Value& v1 = it1->getCurrent();
            
            if (v1.support<it1->mXminsup) {
                it1->nextValue();
                continue;
            }
            
            int cbl = cl;
            itemset[depth-1] = pair<int,string*>(it1->attr,v1.val);
            multiset<Attribute> *children = new multiset<Attribute>;
            
            int mX1 = tail.support;
            double mXminsup1 = minsup * ((mX1>=minrep*tcount)?mX1:minrep*tcount);
            double mXminsup2 = mXminsup1;
            
            for(it2=it1, it2++, end=items->end(); it2!=end; it2++) {
                
                if (!fr[it2->attr]) continue;
                
                Attribute& a = *(fa[it2->attr]);
                
                mXminsup2 = minsup * ((maxtails[a.attr]>=minrepcount)?maxtails[a.attr]:minrepcount);
                
                a.mXminsup = mXminsup2;
                
                it2->reset();
                while(it2->currentIsValid()) {
                    const Value& v2 = it2->getCurrent();
                    if (v2.support<it1->mXminsup) {
                        it2->nextValue();
                        continue;
                    }
                    
                    Value item(v2.val);
                    
                    insert_iterator< vector<int> > res_ins(*(item.transactions), item.transactions->begin());
                    if (depth==1)
                        set_difference(v1.transactions->begin(), v1.transactions->end(), v2.transactions->begin(), v2.transactions->end(), res_ins);
                    else
                        set_difference(v2.transactions->begin(), v2.transactions->end(), v1.transactions->begin(), v1.transactions->end(), res_ins);
                    item.support = v1.support - item.transactions->size();
                    
                    ++candidates;
                    
                    if (item.support==v1.support && a.attrItem->support==it1->attrItem->support) { // closed // pray that no two of same *it2 are put in comb
                        comb[cbl].first = a.attr;
                        comb[cbl].second = item.val;
                        ++cbl;
                    }
                    else if (item.support>=mXminsup1) {
                        a.addValue(item);
                    }
                    
                    it2->nextValue();
                }
                
                a.reset();
                if (!a.values->empty())
                    children->insert(a);
                else {
                    ;//forall i in maxtails do 
                }
                a.values->clear();
                
            } //it2
            
            if ((double)(v1.support) >= (minsup*v1s.support)) {
                print(itemset, depth, comb, cbl, v1.support,v1s.support);
                ++added;
            }
            
            double tmp = grow(children, itemset, depth+1, comb, cbl);
            added += tmp;
            if ((double)(v1.support) >= (minsup*v1s.support) || tmp>0)
                ++extensibleCandidates;
            
            delete children;
            
            it1->nextValue();
        }
        
        map<int,Attribute*>::iterator mit = fa.begin();
        for(;mit!=fa.end();mit++) {
            delete mit->second;
        }
        
    } //it1
    
    return added;
}


void XMiner::print(pair<int,string*> *itemset, int il, pair<int,string*> *comb, int cl, int support, int denom) const
{
    if(out) {
        set<pair<int,string*> > outset;
        for(int i=0; i<cl; ++i) outset.insert(comb[i]);
        for(int j=0; j<il; ++j) outset.insert(itemset[j]);
        for(set<pair<int,string*> >::iterator k=outset.begin(), end=outset.end(); k!=end; k++) {
            if (!k->second)
                fprintf(out, "(%d=?)", k->first);
            else if(*(k->second)=="")
                fprintf(out, "(%d=*)", k->first);
            else
                fprintf(out, "(%d=%s)", k->first, k->second->c_str());
        }
        fprintf(out, " (%d/%d) %f\n", support, denom, (double)support/(double)denom);
    }
}
